@@ -3,7 +3,7 @@ import venusian
 import six
 
 
-ADT_VARIANT_NAME_RE = re.compile('[A-Z][0-9A-Z_]*')
+ADT_VARIANT_NAME_RE = re.compile('^[A-Z][0-9A-Z_]*$')
 
 
 class ADTVariant(object):
@@ -20,6 +20,15 @@ class ADTVariant(object):
         self.name = name
         self.value = value
 
+    def is_primitive_type(self):
+        if isinstance(self.value, six.string_types):
+            primitive_type = True
+        elif isinstance(self.value, six.integer_types):
+            primitive_type = True
+        else:
+            primitive_type = False
+        return primitive_type
+
 
     def __call__(self, case_name):
         """ This method is called when we declare variant cases. For instance:
@@ -76,12 +85,13 @@ class ADTMeta(type):
     def __new__(mcs, class_name, bases, attrs):
         cls = type.__new__(mcs, class_name, bases, attrs)
 
-        variants = set()
+        variants = {}
         for attr_name, value in attrs.items():
             if not ADT_VARIANT_NAME_RE.match(attr_name):
                 continue
-            setattr(cls, attr_name, ADTVariant(variant_of=cls, name=attr_name, value=value))
-            variants.add(attr_name)
+            variant = ADTVariant(variant_of=cls, name=attr_name, value=value)
+            setattr(cls, attr_name, variant)
+            variants[attr_name] = variant
 
         values = {attrs[variant_name]: variant_name for variant_name in variants}
         cls.__adt__ = {
@@ -93,7 +103,9 @@ class ADTMeta(type):
             'cases': {},
             # dict of value => match instances.
             # Used by .match() for O(1) result retrieval
-            'matches': {}
+            'matches': {
+                variant: {} for variant in variants
+            }
         }
         return cls
 
@@ -105,6 +117,9 @@ class adt(object):
     class Mismatch(Exception):
         pass
 
+    class PatternError(Exception):
+        pass
+
     @classmethod
     def values(cls):
         return set(cls.__adt__['values'].keys())
@@ -112,11 +127,22 @@ class adt(object):
     @classmethod
     def match(cls, value):
         """
-        :rtype: dict
+        :rtype: dict or :class:`types.FunctionType`
         """
-        try:
-            variant = cls.__adt__['values'][value]
-        except KeyError:
+        variant = None
+        for variant_name, variant_type in cls.__adt__['variants'].items():
+            if variant_type.is_primitive_type():
+                # We compare primitive types with equality matching
+                if value == variant_type.value:
+                    variant = variant_name
+                    break
+            else:
+                # We compare non-primitive types with type checking
+                if isinstance(value, variant_type.value):
+                    variant = variant_name
+                    break
+
+        if variant is None:
             raise cls.Mismatch(
                 u'Variant value "{value}" is not a part of the type {type}: {values}'.format(
                     value=value,
@@ -128,5 +154,50 @@ class adt(object):
 
         return cls.__adt__['matches'][variant]
 
+    @classmethod
+    def inline_match(cls, **inline_cases):
+        all_cases = set(cls.__adt__['variants'].keys())
+        inline_cases = inline_cases.items()
+        checked_cases = []
+        for variant_name, fun in inline_cases:
+            try:
+                variant = cls.__adt__['variants'][variant_name]
+            except KeyError:
+                raise cls.PatternError(
+                    'Variant {variant} does not belong to the type {type}'.format(
+                        variant=str(variant_name),
+                        type=cls.__adt__['type'],
+                    )
+                )
+            all_cases.remove(variant.name)
+            checked_cases.append((variant, fun))
+
+        if all_cases:
+            raise cls.PatternError(
+                'Inline cases are not exhaustive.\n'
+                'Here is the variant that is not matched: {variant} '.format(
+                    variant=list(all_cases)[0]
+                )
+            )
+
+        def matcher(value):
+            for variant, fun in checked_cases:
+                if variant.is_primitive_type():
+                    if value == variant.value:
+                        return fun
+                else:
+                    if isinstance(value, variant.value):
+                        return fun
+
+            raise cls.Mismatch(
+                u'Variant value "{value}" is not a part of the type {type}: {values}'.format(
+                    value=value,
+                    type=cls.__adt__['type'],
+                    values=u', '.join(['{val} => {var}'.format(val=val, var=var)
+                                       for val, var in cls.__adt__['values'].items()])
+                )
+            )
+        return matcher
+
     def __init__(self):
         raise TypeError('adt is a type, not an instance.')
